#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litepcie.tlp.common import *

# LitePCIeTLPHeaderExtracter64b --------------------------------------------------------------------

class LitePCIeTLPHeaderExtracter64b(Module):
    def __init__(self, endianness):
        self.sink   = sink   = stream.Endpoint(phy_layout(64))
        self.source = source = stream.Endpoint(tlp_raw_layout(64))

        # # #

        first = Signal()
        last  = Signal()
        count = Signal()
        dat   = Signal(64, reset_less=True)
        be    = Signal(64//8, reset_less=True)
        self.sync += \
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be)
            )

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            NextValue(first, 1),
            NextValue(last, 0),
            NextValue(count, 0),
            If(sink.valid, NextState("HEADER"))
        )
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid,
                NextValue(count, count + 1),
                NextValue(source.header[32*0:32*1], source.header[32*2:32*3]),
                NextValue(source.header[32*1:32*2], source.header[32*3:32*4]),
                NextValue(source.header[32*2:32*3], sink.dat[32*0:32*1]),
                NextValue(source.header[32*3:32*4], sink.dat[32*1:32*2]),
                If(count,
                    If(sink.last, NextValue(last, 1)),
                    NextState("COPY")
                )
            )
        )
        fsm.act("COPY",
            source.valid.eq(sink.valid | last),
            source.first.eq(first),
            source.last.eq(sink.last | last),
            If(source.valid & source.ready,
                NextValue(first, 0),
                sink.ready.eq(1 & ~last), # already acked when last is 1
                If(source.last, NextState("IDLE"))
            )
        )
        self.comb += [
            source.dat[32*0:32*1].eq(convert_bytes(dat[32*1:32*2], endianness)),
            source.dat[32*1:32*2].eq(convert_bytes(sink.dat[32*0:32*1], endianness)),
            source.be[4*0:4*1].eq(convert_bits(be[4*1:4*2], endianness)),
            source.be[4*1:4*2].eq(convert_bits(sink.be[4*0:4*1], endianness))
        ]

# LitePCIeTLPHeaderExtracter128b -------------------------------------------------------------------

class LitePCIeTLPHeaderExtracter128b(Module):
    def __init__(self, endianness):
        self.sink   = sink   = stream.Endpoint(phy_layout(128))
        self.source = source = stream.Endpoint(tlp_raw_layout(128))

        # # #

        first = Signal()
        last  = Signal()
        dat   = Signal(128, reset_less=True)
        be    = Signal(128//8, reset_less=True)
        self.sync += \
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be)
            )

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            NextValue(first, 1),
            NextValue(last, 0),
            If(sink.valid,
                NextState("HEADER")
            )
        )
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid,
                NextValue(source.header[32*0:32*1], sink.dat[32*0:32*1]),
                NextValue(source.header[32*1:32*2], sink.dat[32*1:32*2]),
                NextValue(source.header[32*2:32*3], sink.dat[32*2:32*3]),
                NextValue(source.header[32*3:32*4], sink.dat[32*3:32*4]),
                If(sink.last,
                    NextValue(last, 1)
                ),
                NextState("COPY")
            )
        )
        fsm.act("COPY",
            source.valid.eq(sink.valid | last),
            source.first.eq(first),
            source.last.eq(sink.last | last),
            If(source.valid & source.ready,
                NextValue(first, 0),
                sink.ready.eq(1 & ~last), # already acked when last is 1
                If(source.last,
                    NextState("IDLE")
                )
            )
        )
        self.comb += [
            source.dat[32*0:32*1].eq(convert_bytes(dat[32*3:32*4], endianness)),
            source.dat[32*1:32*2].eq(convert_bytes(sink.dat[32*0:32*1], endianness)),
            source.dat[32*2:32*3].eq(convert_bytes(sink.dat[32*1:32*2], endianness)),
            source.dat[32*3:32*4].eq(convert_bytes(sink.dat[32*2:32*3], endianness)),
            source.be[4*0:4*1].eq(convert_bits(be[4*3:4*4], endianness)),
            source.be[4*1:4*2].eq(convert_bits(sink.be[4*0:4*1], endianness)),
            source.be[4*2:4*3].eq(convert_bits(sink.be[4*1:4*2], endianness)),
            source.be[4*1:4*2].eq(convert_bits(sink.be[4*2:4*3], endianness))
        ]

# LitePCIeTLPHeaderExtracter256b -------------------------------------------------------------------

class LitePCIeTLPHeaderExtracter256b(Module):
    def __init__(self, endianness):
        self.sink   = sink   = stream.Endpoint(phy_layout(256))
        self.source = source = stream.Endpoint(tlp_raw_layout(256))

        # # #

        first = Signal()
        last  = Signal()
        dat   = Signal(256, reset_less=True)
        be    = Signal(256//8, reset_less=True)
        self.sync += \
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be)
            )

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            NextValue(first, 1),
            NextValue(last, 0),
            If(sink.valid,
                NextState("HEADER")
            )
        )
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid,
                NextValue(source.header[32*0:32*1], sink.dat[32*0:32*1]),
                NextValue(source.header[32*1:32*2], sink.dat[32*1:32*2]),
                NextValue(source.header[32*2:32*3], sink.dat[32*2:32*3]),
                NextValue(source.header[32*3:32*4], sink.dat[32*3:32*4]),
                If(sink.last,
                    NextValue(last, 1)
                ),
                NextState("COPY")
            )
        )
        fsm.act("COPY",
            source.valid.eq(sink.valid | last),
            source.first.eq(first),
            source.last.eq(sink.last | last),
            If(source.valid & source.ready,
                NextValue(first, 0),
                sink.ready.eq(1 & ~last), # already acked when last is 1
                If(source.last,
                    NextState("IDLE")
                )
            )
        )
        self.comb += [
            source.dat[32*0:32*1].eq(convert_bytes(dat[32*3:32*4], endianness)),
            source.dat[32*1:32*2].eq(convert_bytes(dat[32*4:32*5], endianness)),
            source.dat[32*2:32*3].eq(convert_bytes(dat[32*5:32*6], endianness)),
            source.dat[32*3:32*4].eq(convert_bytes(dat[32*6:32*7], endianness)),
            source.dat[32*4:32*5].eq(convert_bytes(dat[32*7:32*8], endianness)),
            source.dat[32*5:32*6].eq(convert_bytes(sink.dat[32*0:32*1], endianness)),
            source.dat[32*6:32*7].eq(convert_bytes(sink.dat[32*1:32*2], endianness)),
            source.dat[32*7:32*8].eq(convert_bytes(sink.dat[32*2:32*3], endianness)),

            source.be[4*0:4*1].eq(convert_bits(be[4*3:4*4], endianness)),
            source.be[4*1:4*2].eq(convert_bits(be[4*4:4*5], endianness)),
            source.be[4*2:4*3].eq(convert_bits(be[4*5:4*6], endianness)),
            source.be[4*3:4*4].eq(convert_bits(be[4*6:4*7], endianness)),
            source.be[4*4:4*5].eq(convert_bits(be[4*7:4*8], endianness)),
            source.be[4*5:4*6].eq(convert_bits(sink.be[4*0:4*1], endianness)),
            source.be[4*6:4*7].eq(convert_bits(sink.be[4*1:4*2], endianness)),
            source.be[4*7:4*8].eq(convert_bits(sink.be[4*2:4*3], endianness))
        ]

# LitePCIeTLPHeaderExtracter512b -------------------------------------------------------------------

class LitePCIeTLPHeaderExtracter512b(Module):
    def __init__(self, endianness):
        self.sink   = sink   = stream.Endpoint(phy_layout(512))
        self.source = source = stream.Endpoint(tlp_raw_layout(512))

        # # #

        first = Signal()
        last  = Signal()
        dat   = Signal(512, reset_less=True)
        be    = Signal(512//8, reset_less=True)
        self.sync += \
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be)
            )

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            NextValue(first, 1),
            NextValue(last, 0),
            If(sink.valid,
                NextState("HEADER")
            )
        )
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid,
                NextValue(source.header[32*0:32*1], sink.dat[32*0:32*1]),
                NextValue(source.header[32*1:32*2], sink.dat[32*1:32*2]),
                NextValue(source.header[32*2:32*3], sink.dat[32*2:32*3]),
                NextValue(source.header[32*3:32*4], sink.dat[32*3:32*4]),
                If(sink.last,
                    NextValue(last, 1)
                ),
                NextState("COPY")
            )
        )
        fsm.act("COPY",
            source.valid.eq(sink.valid | last),
            source.first.eq(first),
            source.last.eq(sink.last | last),
            If(source.valid & source.ready,
                NextValue(first, 0),
                sink.ready.eq(1 & ~last), # already acked when last is 1
                If(source.last,
                    NextState("IDLE")
                )
            )
        )
        self.comb += [
            source.dat[32*0:32*1].eq(convert_bytes(dat[32*3:32*4], endianness)),
            source.dat[32*1:32*2].eq(convert_bytes(dat[32*4:32*5], endianness)),
            source.dat[32*2:32*3].eq(convert_bytes(dat[32*5:32*6], endianness)),
            source.dat[32*3:32*4].eq(convert_bytes(dat[32*6:32*7], endianness)),
            source.dat[32*4:32*5].eq(convert_bytes(dat[32*7:32*8], endianness)),
            source.dat[32*5:32*6].eq(convert_bytes(dat[32*8:32*9], endianness)),
            source.dat[32*6:32*7].eq(convert_bytes(dat[32*9:32*10], endianness)),
            source.dat[32*7:32*8].eq(convert_bytes(dat[32*10:32*11], endianness)),
            source.dat[32*8:32*9].eq(convert_bytes(dat[32*11:32*12], endianness)),
            source.dat[32*9:32*10].eq(convert_bytes(dat[32*12:32*13], endianness)),
            source.dat[32*10:32*11].eq(convert_bytes(dat[32*13:32*14], endianness)),
            source.dat[32*11:32*12].eq(convert_bytes(dat[32*14:32*15], endianness)),
            source.dat[32*12:32*13].eq(convert_bytes(dat[32*15:32*16], endianness)),
            source.dat[32*13:32*14].eq(convert_bytes(sink.dat[32*0:32*1], endianness)),
            source.dat[32*14:32*15].eq(convert_bytes(sink.dat[32*1:32*2], endianness)),
            source.dat[32*15:32*16].eq(convert_bytes(sink.dat[32*2:32*3], endianness)),


            source.be[4*0:4*1].eq(convert_bits(be[4*3:4*4], endianness)),
            source.be[4*1:4*2].eq(convert_bits(be[4*4:4*5], endianness)),
            source.be[4*2:4*3].eq(convert_bits(be[4*5:4*6], endianness)),
            source.be[4*3:4*4].eq(convert_bits(be[4*6:4*7], endianness)),
            source.be[4*4:4*5].eq(convert_bits(be[4*7:4*8], endianness)),
            source.be[4*5:4*6].eq(convert_bits(be[4*8:4*9], endianness)),
            source.be[4*6:4*7].eq(convert_bits(be[4*9:4*10], endianness)),
            source.be[4*7:4*8].eq(convert_bits(be[4*10:4*11], endianness)),
            source.be[4*8:4*9].eq(convert_bits(be[4*11:4*12], endianness)),
            source.be[4*9:4*10].eq(convert_bits(be[4*12:4*13], endianness)),
            source.be[4*10:4*11].eq(convert_bits(be[4*13:4*14], endianness)),
            source.be[4*11:4*12].eq(convert_bits(be[4*14:4*15], endianness)),
            source.be[4*12:4*13].eq(convert_bits(be[4*15:4*16], endianness)),
            source.be[4*13:4*14].eq(convert_bits(sink.be[4*0:4*1], endianness)),
            source.be[4*14:4*15].eq(convert_bits(sink.be[4*1:4*2], endianness)),
            source.be[4*15:4*16].eq(convert_bits(sink.be[4*2:4*3], endianness)),
        ]

# LitePCIeTLPDepacketizer --------------------------------------------------------------------------

class LitePCIeTLPDepacketizer(Module):
    def __init__(self, data_width, endianness, address_mask=0):
        self.sink       = stream.Endpoint(phy_layout(data_width))
        self.req_source = stream.Endpoint(request_layout(data_width))
        self.cmp_source = stream.Endpoint(completion_layout(data_width))

        # # #

        # Extract raw header -----------------------------------------------------------------------
        header_extracter_cls = {
             64 : LitePCIeTLPHeaderExtracter64b,
            128 : LitePCIeTLPHeaderExtracter128b,
            256 : LitePCIeTLPHeaderExtracter256b,
            512 : LitePCIeTLPHeaderExtracter512b,
        }
        header_extracter = header_extracter_cls[data_width](endianness)
        self.submodules += header_extracter
        self.comb += self.sink.connect(header_extracter.sink)
        header = header_extracter.source.header

        # Dispatch data according to fmt/type ------------------------------------------------------
        dispatch_source = stream.Endpoint(tlp_common_layout(data_width))
        dispatch_sinks  = [stream.Endpoint(tlp_common_layout(data_width)) for i in range(2)]

        self.comb += [
            dispatch_source.valid.eq(header_extracter.source.valid),
            header_extracter.source.ready.eq(dispatch_source.ready),
            dispatch_source.first.eq(header_extracter.source.first),
            dispatch_source.last.eq(header_extracter.source.last),
            dispatch_source.dat.eq(header_extracter.source.dat),
            dispatch_source.be.eq(header_extracter.source.be),
            tlp_common_header.decode(header, dispatch_source)
        ]

        self.submodules.dispatcher = Dispatcher(dispatch_source, dispatch_sinks)

        fmt_type = Cat(dispatch_source.type, dispatch_source.fmt)
        self.comb += [
            If((fmt_type == fmt_type_dict["mem_rd32"]) | (fmt_type == fmt_type_dict["mem_wr32"]),
                self.dispatcher.sel.eq(0),
            ).Elif((fmt_type == fmt_type_dict["cpld"]) | (fmt_type == fmt_type_dict["cpl"]),
                self.dispatcher.sel.eq(1)
            )
        ]

        # Decode TLP request and format local request ----------------------------------------------
        self.tlp_req = tlp_req = stream.Endpoint(tlp_request_layout(data_width))
        self.comb += dispatch_sinks[0].connect(tlp_req)
        self.comb += tlp_request_header.decode(header, tlp_req)

        req_source = self.req_source
        self.comb += [
            req_source.valid.eq(tlp_req.valid),
            req_source.we.eq(tlp_req.valid &
                (Cat(tlp_req.type, tlp_req.fmt) == fmt_type_dict["mem_wr32"])),
            tlp_req.ready.eq(req_source.ready),
            req_source.first.eq(tlp_req.first),
            req_source.last.eq(tlp_req.last),
            req_source.adr.eq(tlp_req.address & (~address_mask)),
            req_source.len.eq(tlp_req.length),
            req_source.req_id.eq(tlp_req.requester_id),
            req_source.tag.eq(tlp_req.tag),
            req_source.dat.eq(tlp_req.dat)
        ]

        # Decode TLP completion and format local completion ----------------------------------------
        self.tlp_cmp = tlp_cmp = stream.Endpoint(tlp_completion_layout(data_width))
        self.comb += dispatch_sinks[1].connect(tlp_cmp)
        self.comb += tlp_completion_header.decode(header, tlp_cmp)

        cmp_source = self.cmp_source
        self.comb += [
            cmp_source.valid.eq(tlp_cmp.valid),
            tlp_cmp.ready.eq(cmp_source.ready),
            cmp_source.first.eq(tlp_cmp.first),
            cmp_source.last.eq(tlp_cmp.last),
            cmp_source.len.eq(tlp_cmp.length),
            cmp_source.end.eq(tlp_cmp.length == (tlp_cmp.byte_count[2:])),
            cmp_source.adr.eq(tlp_cmp.lower_address),
            cmp_source.req_id.eq(tlp_cmp.requester_id),
            cmp_source.cmp_id.eq(tlp_cmp.completer_id),
            cmp_source.err.eq(tlp_cmp.status != 0),
            cmp_source.tag.eq(tlp_cmp.tag),
            cmp_source.dat.eq(tlp_cmp.dat)
        ]



#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *
from migen.genlib.misc import chooser

from litepcie.tlp.common import *

# LitePCIeTLPHeaderInserter64b ---------------------------------------------------------------------

class LitePCIeTLPHeaderInserter64b(Module):
    def __init__(self, endianness):
        self.sink   = sink   = stream.Endpoint(tlp_raw_layout(64))
        self.source = source = stream.Endpoint(phy_layout(64))

        # # #

        dat  = Signal(64, reset_less=True)
        last = Signal(reset_less=True)
        self.sync += \
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                last.eq(sink.last)
            )

        self.submodules.fsm = fsm = FSM(reset_state="HEADER1")
        fsm.act("HEADER1",
            sink.ready.eq(1),
            If(sink.first,
                sink.ready.eq(0),
                source.valid.eq(sink.valid),
                source.first.eq(sink.first),
                source.last.eq(0),
                # 32 lsbs
                source.dat[32*0:32*1].eq(sink.header[32*0:32*1]),
                source.be[4*0:4*1].eq(0xf),
                # 32 msbs
                source.dat[32*1:32*2].eq(sink.header[32*1:32*2]),
                source.be[4*1:4*2].eq(0xf),
                If(source.valid & source.ready,
                    NextState("HEADER2"),
                )
            )
        )
        fsm.act("HEADER2",
            source.valid.eq(1),
            source.last.eq(sink.last),
            # 32 lsbs
            source.dat[32*0:32*1].eq(sink.header[32*2:32*3]),
            source.be[4*0:4*1].eq(0xf),
            # 32 msbs
            source.dat[32*1:32*2].eq(convert_bytes(sink.dat[32*0:32*1], endianness)),
            source.be[4*1:4*2].eq(convert_bits(sink.be[:4], endianness)),
            If(source.valid & source.ready,
                sink.ready.eq(1),
                If(source.last,
                    NextState("HEADER1")
                ).Else(
                    NextState("DATA")
                )
            )
        )
        fsm.act("DATA",
            source.valid.eq(sink.valid | last),
            source.last.eq(last),
            # 32 msbs
            source.dat[32*0:32*1].eq(convert_bytes(dat[32*1:32*2], endianness)),
			source.be[4*0:4*1].eq(0xf),
            # 32 lsbs
            source.dat[32*1:32*2].eq(convert_bytes(sink.dat[32*0:32*1], endianness)),
            If(last,
                source.be[4*1:4*2].eq(0x0)
            ).Else(
                source.be[4*1:4*2].eq(0xf)
            ),
            If(source.valid & source.ready,
                sink.ready.eq(~last),
                If(source.last,
                    NextState("HEADER1")
                )
            )
        )

# LitePCIeTLPHeaderInserter128b --------------------------------------------------------------------

class LitePCIeTLPHeaderInserter128b(Module):
    def __init__(self, endianness):
        self.sink   = sink   = stream.Endpoint(tlp_raw_layout(128))
        self.source = source = stream.Endpoint(phy_layout(128))

        # # #

        dat  = Signal(128, reset_less=True)
        last = Signal(reset_less=True)
        self.sync += \
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                last.eq(sink.last)
            )

        self.submodules.fsm = fsm = FSM(reset_state="HEADER")
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid & sink.first,
                sink.ready.eq(0),
                source.valid.eq(1),
                source.first.eq(1),
                source.last.eq(sink.last),
                # 32 l lsbs
                source.dat[32*0:32*1].eq(sink.header[32*0:32*1]),
                source.be[4*0:4*1].eq(0xf),
                # 32 l msbs
                source.dat[32*1:32*2].eq(sink.header[32*1:32*2]),
                source.be[4*1:4*2].eq(0xf),
                # 32 h lsbs
                source.dat[32*2:32*3].eq(sink.header[32*2:32*3]),
                source.be[4*2:4*3].eq(0xf),
                # 32 h msbs
                source.dat[32*3:32*4].eq(convert_bytes(sink.dat[32*0:32*1], endianness)),
                source.be[4*3:4*4].eq(convert_bits(sink.be[:4], endianness)),
                If(source.valid & source.ready,
                    sink.ready.eq(1),
                    If(~source.last,
                        NextState("DATA"),
                    )
                )
            )
        )
        fsm.act("DATA",
            source.valid.eq(sink.valid | last),
            source.last.eq(last),
            # 32 l lsbs
            source.dat[32*0:32*1].eq(convert_bytes(dat[32*1:32*2], endianness)),
            source.be[4*0:4*1].eq(0xf),
            # 32 l msbs
            source.dat[32*1:32*2].eq(convert_bytes(dat[32*2:32*3], endianness)),
            source.be[4*1:4*2].eq(0xf),
            # 32 h lsbs
            source.dat[32*2:32*3].eq(convert_bytes(dat[32*3:2*32*4], endianness)),
            source.be[4*2:4*3].eq(0xf),
            # 32 h msbs
            source.dat[32*3:32*4].eq(convert_bytes(sink.dat[32*0:32*1], endianness)),
            If(last,
                source.be[4*3:4*4].eq(0x0)
            ).Else(
                source.be[4*3:4*4].eq(0xf)
            ),
            If(source.valid & source.ready,
                sink.ready.eq(~last),
                If(source.last,
                    NextState("HEADER")
                )
            )
        )

# LitePCIeTLPHeaderInserter256b --------------------------------------------------------------------

class LitePCIeTLPHeaderInserter256b(Module):
    def __init__(self, endianness):
        self.sink   = sink   = stream.Endpoint(tlp_raw_layout(256))
        self.source = source = stream.Endpoint(phy_layout(256))

        # # #

        dat  = Signal(256, reset_less=True)
        be  = Signal(256//8, reset_less=True)
        last = Signal(reset_less=True)
        self.sync += \
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be),
                last.eq(sink.last)
            )

        self.submodules.fsm = fsm = FSM(reset_state="HEADER")
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid & sink.first,
                sink.ready.eq(0),
                source.valid.eq(1),
                source.first.eq(1),
                source.last.eq(sink.last),

                source.dat[32*0:32*1].eq(sink.header[32*0:32*1]),
                source.dat[32*1:32*2].eq(sink.header[32*1:32*2]),
                source.dat[32*2:32*3].eq(sink.header[32*2:32*3]),
                source.dat[32*3:32*4].eq(convert_bytes(sink.dat[32*0:32*1], endianness)),
                source.dat[32*4:32*5].eq(convert_bytes(sink.dat[32*1:32*2], endianness)),
                source.dat[32*5:32*6].eq(convert_bytes(sink.dat[32*2:32*3], endianness)),
                source.dat[32*6:32*7].eq(convert_bytes(sink.dat[32*3:32*4], endianness)),
                source.dat[32*7:32*8].eq(convert_bytes(sink.dat[32*4:32*5], endianness)),

                source.be[4*0:4*1].eq(0xf),
                source.be[4*1:4*2].eq(0xf),
                source.be[4*2:4*3].eq(0xf),
                source.be[4*3:4*4].eq(convert_bits(sink.be[4*0:4*1], endianness)),
                source.be[4*4:4*5].eq(convert_bits(sink.be[4*1:4*2], endianness)),
                source.be[4*5:4*6].eq(convert_bits(sink.be[4*2:4*3], endianness)),
                source.be[4*6:4*7].eq(convert_bits(sink.be[4*3:4*4], endianness)),
                source.be[4*7:4*8].eq(convert_bits(sink.be[4*4:4*5], endianness)),

                If(source.valid & source.ready,
                    sink.ready.eq(1),
                    If(~source.last,
                        NextState("DATA"),
                    )
                )
            )
        )
        fsm.act("DATA",
            source.valid.eq(sink.valid | last),
            source.last.eq(last),

            source.dat[32*0:32*1].eq(convert_bytes(dat[32*5:32*6], endianness)),
            source.dat[32*1:32*2].eq(convert_bytes(dat[32*6:32*7], endianness)),
            source.dat[32*2:32*3].eq(convert_bytes(dat[32*7:32*8], endianness)),
            source.dat[32*3:32*4].eq(convert_bytes(sink.dat[32*0:32*1], endianness)),
            source.dat[32*4:32*5].eq(convert_bytes(sink.dat[32*1:32*2], endianness)),
            source.dat[32*5:32*6].eq(convert_bytes(sink.dat[32*2:32*3], endianness)),
            source.dat[32*6:32*7].eq(convert_bytes(sink.dat[32*3:32*4], endianness)),
            source.dat[32*7:32*8].eq(convert_bytes(sink.dat[32*4:32*5], endianness)),

            source.be[4*0:4*1].eq(convert_bits(be[4*5:4*6], endianness)),
            source.be[4*1:4*2].eq(convert_bits(be[4*6:4*7], endianness)),
            source.be[4*2:4*3].eq(convert_bits(be[4*7:4*8], endianness)),
            If(last,
                source.be[4*3:4*8].eq(0x0)
            ).Else(
                source.be[4*3:4*4].eq(convert_bits(sink.be[4*0:4*1], endianness)),
                source.be[4*4:4*5].eq(convert_bits(sink.be[4*1:4*2], endianness)),
                source.be[4*5:4*6].eq(convert_bits(sink.be[4*2:4*3], endianness)),
                source.be[4*6:4*7].eq(convert_bits(sink.be[4*3:4*4], endianness)),
                source.be[4*7:4*8].eq(convert_bits(sink.be[4*4:4*5], endianness)),
            ),
            If(source.valid & source.ready,
                sink.ready.eq(~last),
                If(source.last,
                    NextState("HEADER")
                )
            )
        )

# LitePCIeTLPHeaderInserter512b --------------------------------------------------------------------

class LitePCIeTLPHeaderInserter512b(Module):
    def __init__(self, endianness):
        self.sink   = sink   = stream.Endpoint(tlp_raw_layout(512))
        self.source = source = stream.Endpoint(phy_layout(512))

        # # #

        dat  = Signal(512, reset_less=True)
        be  = Signal(512//8, reset_less=True)
        last = Signal(reset_less=True)
        self.sync += \
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be),
                last.eq(sink.last)
            )

        self.submodules.fsm = fsm = FSM(reset_state="HEADER")
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid & sink.first,
                sink.ready.eq(0),
                source.valid.eq(1),
                source.first.eq(1),
                source.last.eq(sink.last),

                source.dat[32*0:32*1].eq(sink.header[32*0:32*1]),
                source.dat[32*1:32*2].eq(sink.header[32*1:32*2]),
                source.dat[32*2:32*3].eq(sink.header[32*2:32*3]),
                source.dat[32*3:32*4].eq(convert_bytes(sink.dat[32*0:32*1], endianness)),
                source.dat[32*4:32*5].eq(convert_bytes(sink.dat[32*1:32*2], endianness)),
                source.dat[32*5:32*6].eq(convert_bytes(sink.dat[32*2:32*3], endianness)),
                source.dat[32*6:32*7].eq(convert_bytes(sink.dat[32*3:32*4], endianness)),
                source.dat[32*7:32*8].eq(convert_bytes(sink.dat[32*4:32*5], endianness)),
                source.dat[32*8:32*9].eq(convert_bytes(sink.dat[32*5:32*6], endianness)),
                source.dat[32*9:32*10].eq(convert_bytes(sink.dat[32*6:32*7], endianness)),
                source.dat[32*10:32*11].eq(convert_bytes(sink.dat[32*7:32*8], endianness)),
                source.dat[32*11:32*12].eq(convert_bytes(sink.dat[32*8:32*9], endianness)),
                source.dat[32*12:32*13].eq(convert_bytes(sink.dat[32*9:32*10], endianness)),
                source.dat[32*13:32*14].eq(convert_bytes(sink.dat[32*10:32*11], endianness)),
                source.dat[32*14:32*15].eq(convert_bytes(sink.dat[32*11:32*12], endianness)),
                source.dat[32*15:32*16].eq(convert_bytes(sink.dat[32*12:32*13], endianness)),

                source.be[4*0:4*1].eq(0xf),
                source.be[4*1:4*2].eq(0xf),
                source.be[4*2:4*3].eq(0xf),
                source.be[4*3:4*4].eq(convert_bits(sink.be[4*0:4*1], endianness)),
                source.be[4*4:4*5].eq(convert_bits(sink.be[4*1:4*2], endianness)),
                source.be[4*5:4*6].eq(convert_bits(sink.be[4*2:4*3], endianness)),
                source.be[4*6:4*7].eq(convert_bits(sink.be[4*3:4*4], endianness)),
                source.be[4*7:4*8].eq(convert_bits(sink.be[4*4:4*5], endianness)),
                source.be[4*8:4*9].eq(convert_bits(sink.be[4*5:4*6], endianness)),
                source.be[4*9:4*10].eq(convert_bits(sink.be[4*6:4*7], endianness)),
                source.be[4*10:4*11].eq(convert_bits(sink.be[4*7:4*8], endianness)),
                source.be[4*11:4*12].eq(convert_bits(sink.be[4*8:4*9], endianness)),
                source.be[4*12:4*13].eq(convert_bits(sink.be[4*9:4*10], endianness)),
                source.be[4*13:4*14].eq(convert_bits(sink.be[4*10:4*11], endianness)),
                source.be[4*14:4*15].eq(convert_bits(sink.be[4*11:4*12], endianness)),
                source.be[4*15:4*16].eq(convert_bits(sink.be[4*12:4*13], endianness)),

                If(source.valid & source.ready,
                    sink.ready.eq(1),
                    If(~source.last,
                        NextState("DATA"),
                    )
                )
            )
        )
        fsm.act("DATA",
            source.valid.eq(sink.valid | last),
            source.last.eq(last),

            source.dat[32*0:32*1].eq(convert_bytes(dat[32*13:32*14], endianness)),
            source.dat[32*1:32*2].eq(convert_bytes(dat[32*14:32*15], endianness)),
            source.dat[32*2:32*3].eq(convert_bytes(dat[32*15:32*16], endianness)),
            source.dat[32*3:32*4].eq(convert_bytes(sink.dat[32*0:32*1], endianness)),
            source.dat[32*4:32*5].eq(convert_bytes(sink.dat[32*1:32*2], endianness)),
            source.dat[32*5:32*6].eq(convert_bytes(sink.dat[32*2:32*3], endianness)),
            source.dat[32*6:32*7].eq(convert_bytes(sink.dat[32*3:32*4], endianness)),
            source.dat[32*7:32*8].eq(convert_bytes(sink.dat[32*4:32*5], endianness)),
            source.dat[32*8:32*9].eq(convert_bytes(sink.dat[32*5:32*6], endianness)),
            source.dat[32*9:32*10].eq(convert_bytes(sink.dat[32*6:32*7], endianness)),
            source.dat[32*10:32*11].eq(convert_bytes(sink.dat[32*7:32*8], endianness)),
            source.dat[32*11:32*12].eq(convert_bytes(sink.dat[32*8:32*9], endianness)),
            source.dat[32*12:32*13].eq(convert_bytes(sink.dat[32*9:32*10], endianness)),
            source.dat[32*13:32*14].eq(convert_bytes(sink.dat[32*10:32*11], endianness)),
            source.dat[32*14:32*15].eq(convert_bytes(sink.dat[32*11:32*12], endianness)),
            source.dat[32*15:32*16].eq(convert_bytes(sink.dat[32*12:32*13], endianness)),

            source.be[4*0:4*1].eq(convert_bits(be[4*13:4*14], endianness)),
            source.be[4*1:4*2].eq(convert_bits(be[4*14:4*15], endianness)),
            source.be[4*2:4*3].eq(convert_bits(be[4*15:4*16], endianness)),
            If(last,
                source.be[4*3:4*16].eq(0x0)
            ).Else(
                source.be[4*3:4*4].eq(convert_bits(sink.be[4*0:4*1], endianness)),
                source.be[4*4:4*5].eq(convert_bits(sink.be[4*1:4*2], endianness)),
                source.be[4*5:4*6].eq(convert_bits(sink.be[4*2:4*3], endianness)),
                source.be[4*6:4*7].eq(convert_bits(sink.be[4*3:4*4], endianness)),
                source.be[4*7:4*8].eq(convert_bits(sink.be[4*4:4*5], endianness)),
                source.be[4*8:4*9].eq(convert_bits(sink.be[4*5:4*6], endianness)),
                source.be[4*9:4*10].eq(convert_bits(sink.be[4*6:4*7], endianness)),
                source.be[4*10:4*11].eq(convert_bits(sink.be[4*7:4*8], endianness)),
                source.be[4*11:4*12].eq(convert_bits(sink.be[4*8:4*9], endianness)),
                source.be[4*12:4*13].eq(convert_bits(sink.be[4*9:4*10], endianness)),
                source.be[4*13:4*14].eq(convert_bits(sink.be[4*10:4*11], endianness)),
                source.be[4*14:4*15].eq(convert_bits(sink.be[4*11:4*12], endianness)),
                source.be[4*15:4*16].eq(convert_bits(sink.be[4*12:4*13], endianness)),
            ),
            If(source.valid & source.ready,
                sink.ready.eq(~last),
                If(source.last,
                    NextState("HEADER")
                )
            )
        )

# LitePCIeTLPPacketizer ----------------------------------------------------------------------------

class LitePCIeTLPPacketizer(Module):
    def __init__(self, data_width, endianness):
        self.req_sink = req_sink = stream.Endpoint(request_layout(data_width))
        self.cmp_sink = cmp_sink = stream.Endpoint(completion_layout(data_width))
        self.source   = stream.Endpoint(phy_layout(data_width))

        # # #

        # Format TLP request and encode it ---------------------------------------------------------
        self.tlp_req = tlp_req = stream.Endpoint(tlp_request_layout(data_width))
        self.comb += [
            tlp_req.valid.eq(req_sink.valid),
            req_sink.ready.eq(tlp_req.ready),
            tlp_req.first.eq(req_sink.first),
            tlp_req.last.eq(req_sink.last),

            If(req_sink.we,
                Cat(tlp_req.type, tlp_req.fmt).eq(fmt_type_dict["mem_wr32"])
            ).Else(
                Cat(tlp_req.type, tlp_req.fmt).eq(fmt_type_dict["mem_rd32"])
            ),

            tlp_req.tc.eq(0),
            tlp_req.td.eq(0),
            tlp_req.ep.eq(0),
            tlp_req.attr.eq(0),
            tlp_req.length.eq(req_sink.len),

            tlp_req.requester_id.eq(req_sink.req_id),
            tlp_req.tag.eq(req_sink.tag),
            If(req_sink.len > 1,
                tlp_req.last_be.eq(0xf)
            ).Else(
                tlp_req.last_be.eq(0x0)
            ),
            tlp_req.first_be.eq(0xf),
            tlp_req.address.eq(req_sink.adr),

            tlp_req.dat.eq(req_sink.dat),
            If(req_sink.we,
                tlp_req.be.eq(2**(data_width//8)-1)
            ).Else(
                tlp_req.be.eq(0x00)
            )
        ]

        tlp_raw_req = stream.Endpoint(tlp_raw_layout(data_width))
        self.comb += [
            tlp_raw_req.valid.eq(tlp_req.valid),
            tlp_req.ready.eq(tlp_raw_req.ready),
            tlp_raw_req.first.eq(tlp_req.first),
            tlp_raw_req.last.eq(tlp_req.last),
            tlp_request_header.encode(tlp_req, tlp_raw_req.header),
            tlp_raw_req.dat.eq(tlp_req.dat),
            tlp_raw_req.be.eq(tlp_req.be)
        ]

        # Format TLP completion and encode it ------------------------------------------------------
        self.tlp_cmp = tlp_cmp = stream.Endpoint(tlp_completion_layout(data_width))
        self.comb += [
            tlp_cmp.valid.eq(cmp_sink.valid),
            cmp_sink.ready.eq(tlp_cmp.ready),
            tlp_cmp.first.eq(cmp_sink.first),
            tlp_cmp.last.eq(cmp_sink.last),

            tlp_cmp.tc.eq(0),
            tlp_cmp.td.eq(0),
            tlp_cmp.ep.eq(0),
            tlp_cmp.attr.eq(0),
            tlp_cmp.length.eq(cmp_sink.len),

            tlp_cmp.completer_id.eq(cmp_sink.cmp_id),
            If(cmp_sink.err,
                Cat(tlp_cmp.type, tlp_cmp.fmt).eq(fmt_type_dict["cpl"]),
                tlp_cmp.status.eq(cpl_dict["ur"])
            ).Else(
                Cat(tlp_cmp.type, tlp_cmp.fmt).eq(fmt_type_dict["cpld"]),
                tlp_cmp.status.eq(cpl_dict["sc"])
            ),
            tlp_cmp.bcm.eq(0),
            tlp_cmp.byte_count.eq(cmp_sink.len*4),

            tlp_cmp.requester_id.eq(cmp_sink.req_id),
            tlp_cmp.tag.eq(cmp_sink.tag),
            tlp_cmp.lower_address.eq(cmp_sink.adr),

            tlp_cmp.dat.eq(cmp_sink.dat),
            If(cmp_sink.last & cmp_sink.first,
                tlp_cmp.be.eq(0xf)
            ).Else(
                tlp_cmp.be.eq(2**(data_width//8)-1)
            ),
        ]

        tlp_raw_cmp = stream.Endpoint(tlp_raw_layout(data_width))
        self.comb += [
            tlp_raw_cmp.valid.eq(tlp_cmp.valid),
            tlp_cmp.ready.eq(tlp_raw_cmp.ready),
            tlp_raw_cmp.first.eq(tlp_cmp.first),
            tlp_raw_cmp.last.eq(tlp_cmp.last),
            tlp_completion_header.encode(tlp_cmp, tlp_raw_cmp.header),
            tlp_raw_cmp.dat.eq(tlp_cmp.dat),
            tlp_raw_cmp.be.eq(tlp_cmp.be)
        ]

        # Arbitrate --------------------------------------------------------------------------------
        tlp_raw = stream.Endpoint(tlp_raw_layout(data_width))
        self.submodules.arbitrer = Arbiter([tlp_raw_req, tlp_raw_cmp], tlp_raw)

        # Insert header ----------------------------------------------------------------------------
        header_inserter_cls = {
            64 : LitePCIeTLPHeaderInserter64b,
           128 : LitePCIeTLPHeaderInserter128b,
           256 : LitePCIeTLPHeaderInserter256b,
           512 : LitePCIeTLPHeaderInserter512b,
        }
        header_inserter = header_inserter_cls[data_width](endianness)
        self.submodules += header_inserter
        self.comb += [
            tlp_raw.connect(header_inserter.sink),
            header_inserter.source.connect(self.source)
        ]



#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litepcie.common import *

# Constants ----------------------------------------------------------------------------------------

fmt_type_dict = {
    "mem_rd32": 0b0000000,
    "mem_wr32": 0b1000000,
    "mem_rd64": 0b0100000,
    "mem_wr64": 0b1100000,

    "cpld":     0b1001010,
    "cpl":      0b0001010
}

cpl_dict = {
    "sc":  0b000,
    "ur":  0b001,
    "crs": 0b010,
    "ca":  0b011
}

max_payload_size = 512
max_request_size = 512

# Headers ------------------------------------------------------------------------------------------

tlp_common_header_length = 16
tlp_common_header_fields = {
    "fmt":  HeaderField(0*4, 29, 2),
    "type": HeaderField(0*4, 24, 5),
}
tlp_common_header = Header(tlp_common_header_fields,
                           tlp_common_header_length,
                           swap_field_bytes=False)


tlp_request_header_length = 16
tlp_request_header_fields = {
    "fmt":          HeaderField(0*4, 29,  2),
    "type":         HeaderField(0*4, 24,  5),
    "tc":           HeaderField(0*4, 20,  3),
    "td":           HeaderField(0*4, 15,  1),
    "ep":           HeaderField(0*4, 14,  1),
    "attr":         HeaderField(0*4, 12,  2),
    "length":       HeaderField(0*4,  0, 10),

    "requester_id": HeaderField(1*4, 16, 16),
    "tag":          HeaderField(1*4,  8,  8),
    "last_be":      HeaderField(1*4,  4,  4),
    "first_be":     HeaderField(1*4,  0,  4),

    "address":      HeaderField(2*4,  0, 32),
}
tlp_request_header = Header(tlp_request_header_fields,
                            tlp_request_header_length,
                            swap_field_bytes=False)


tlp_completion_header_length = 16
tlp_completion_header_fields = {
    "fmt":           HeaderField(0*4, 29,  2),
    "type":          HeaderField(0*4, 24,  5),
    "tc":            HeaderField(0*4, 20,  3),
    "td":            HeaderField(0*4, 15,  1),
    "ep":            HeaderField(0*4, 14,  1),
    "attr":          HeaderField(0*4, 12,  2),
    "length":        HeaderField(0*4,  0, 10),

    "completer_id":  HeaderField(1*4, 16, 16),
    "status":        HeaderField(1*4, 13,  3),
    "bcm":           HeaderField(1*4, 12,  1),
    "byte_count":    HeaderField(1*4,  0, 12),

    "requester_id":  HeaderField(2*4, 16, 16),
    "tag":           HeaderField(2*4,  8,  8),
    "lower_address": HeaderField(2*4,  0,  7),
}
tlp_completion_header = Header(tlp_completion_header_fields,
                               tlp_completion_header_length,
                               swap_field_bytes=False)

# Helpers ------------------------------------------------------------------------------------------
def convert_bytes(s, endianness="big"):
    return reverse_bytes(s) if endianness == "big" else s

def convert_bits(s, endianness="big"):
    return reverse_bits(s) if endianness == "big" else s

# Layouts ------------------------------------------------------------------------------------------
def tlp_raw_layout(data_width):
    layout = [
        ("header", 4*32),
        ("dat",    data_width),
        ("be",     data_width//8)
    ]
    return EndpointDescription(layout)


def tlp_common_layout(data_width):
    layout = tlp_common_header.get_layout() + [
        ("dat", data_width),
        ("be",  data_width//8)
    ]
    return EndpointDescription(layout)


def tlp_request_layout(data_width):
    layout = tlp_request_header.get_layout() + [
        ("dat", data_width),
        ("be",  data_width//8)
    ]
    return EndpointDescription(layout)


def tlp_completion_layout(data_width):
    layout = tlp_completion_header.get_layout() + [
        ("dat", data_width),
        ("be",  data_width//8)
    ]
    return EndpointDescription(layout)



#!/usr/bin/env python3

#
# This file is part of LitePCIe.
#
# Copyright (c) 2019-2020 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

"""
LitePCIe standalone core generator

LitePCIe aims to be directly used as a python package when the SoC is created using LiteX. However,
for some use cases it could be interesting to generate a standalone verilog file of the core:
- integration of the core in a SoC using a more traditional flow.
- need to version/package the core.
- avoid Migen/LiteX dependencies.
- etc...

The standalone core is generated from a YAML configuration file that allows the user to generate
easily a custom configuration of the core.

Current version of the generator is limited to:
- Xilinx 7-Series.
- Xilinx Ultrascale.
- Altera Cyclone V.
"""

import yaml
import argparse
import subprocess

from migen import *
from migen.genlib.resetsync import AsyncResetSynchronizer
from migen.genlib.misc import WaitTimer

from litex.soc.cores.clock import *
from litex.soc.interconnect.csr import *
from litex.soc.interconnect import wishbone
from litex.soc.interconnect.axi import *
from litex.soc.integration.soc_core import *
from litex.soc.integration.builder import *

from litepcie.core import LitePCIeEndpoint, LitePCIeMSI, LitePCIeMSIMultiVector, LitePCIeMSIX
from litepcie.frontend.dma import LitePCIeDMA
from litepcie.frontend.wishbone import LitePCIeWishboneMaster, LitePCIeWishboneSlave
from litepcie.frontend.axi import LitePCIeAXISlave
from litepcie.software import generate_litepcie_software_headers

from litex.build.generic_platform import *

# IOs/Interfaces -----------------------------------------------------------------------------------

def get_clkin_ios():
    return [
        # clk / rst
        ("clk", 0, Pins(1)),
        ("rst", 0, Pins(1))
    ]

def get_clkout_ios():
    return [
        # clk / rst
        ("clk125", 0, Pins(1)),
        ("rst125", 0, Pins(1))
    ]

def get_pcie_ios(phy_lanes=4):
    return [
        ("pcie", 0,
            Subsignal("rst_n", Pins(1)),
            Subsignal("clk_p", Pins(1)),
            Subsignal("clk_n", Pins(1)),
            Subsignal("rx_p",  Pins(phy_lanes)),
            Subsignal("rx_n",  Pins(phy_lanes)),
            Subsignal("tx_p",  Pins(phy_lanes)),
            Subsignal("tx_n",  Pins(phy_lanes)),
        ),
    ]

def get_axi_dma_ios(_id, dw):
    return [
        ("dma{}_writer_axi".format(_id), 0,
            Subsignal("tvalid", Pins(1)),
            Subsignal("tready", Pins(1)),
            Subsignal("tlast",  Pins(1)),
            Subsignal("tdata",  Pins(dw)),
            Subsignal("tuser",  Pins(1)), # first
        ),
        ("dma{}_reader_axi".format(_id), 0,
            Subsignal("tvalid", Pins(1)),
            Subsignal("tready", Pins(1)),
            Subsignal("tlast",  Pins(1)),
            Subsignal("tdata",  Pins(dw)),
            Subsignal("tuser",  Pins(1)), # first
        ),
    ]

def get_msi_irqs_ios(width=16):
    return [("msi_irqs", 0, Pins(width))]

# CRG ----------------------------------------------------------------------------------------------

class LitePCIeCRG(Module):
    def __init__(self, platform, sys_clk_freq, clk_external):
        self.clock_domains.cd_sys = ClockDomain()

        # # #

        if clk_external:
            platform.add_extension(get_clkin_ios())
            self.comb += self.cd_sys.clk.eq(platform.request("clk"))
            self.specials += AsyncResetSynchronizer(self.cd_sys, platform.request("rst"))
        else:
            platform.add_extension(get_clkout_ios())
            self.comb += self.cd_sys.clk.eq(ClockSignal("pcie"))
            self.comb += self.cd_sys.rst.eq(ResetSignal("pcie"))
            self.comb += platform.request("clk125").eq(ClockSignal("pcie"))
            self.comb += platform.request("rst125").eq(ResetSignal("pcie"))

# Core ---------------------------------------------------------------------------------------------

class LitePCIeCore(SoCMini):
    SoCMini.mem_map["csr"] = 0x00000000
    SoCMini.csr_map = {
        "ctrl":           0,
        "crg" :           1,
        "pcie_phy":       2,
        "pcie_msi":       3,
        "pcie_msi_table": 4,
    }
    def __init__(self, platform, core_config):
        platform.add_extension(get_pcie_ios(core_config["phy_lanes"]))
        for i in range(core_config["dma_channels"]):
            platform.add_extension(get_axi_dma_ios(i, core_config["phy_data_width"]))
        platform.add_extension(get_msi_irqs_ios(width=core_config["msi_irqs"]))
        sys_clk_freq = float(core_config.get("clk_freq", 125e6))

        # SoCMini ----------------------------------------------------------------------------------
        SoCMini.__init__(self, platform, clk_freq=sys_clk_freq,
            csr_data_width = 32,
            csr_ordering   = core_config.get("csr_ordering", "big"),
            ident          = "LitePCIe standalone core",
            ident_version  = True
        )

        # CRG --------------------------------------------------------------------------------------
        clk_external = core_config.get("clk_external", False)
        self.submodules.crg = LitePCIeCRG(platform, sys_clk_freq, clk_external)

        # PCIe PHY ---------------------------------------------------------------------------------
        self.submodules.pcie_phy = core_config["phy"](platform, platform.request("pcie"),
            pcie_data_width = core_config.get("phy_pcie_data_width", 64),
            data_width      = core_config["phy_data_width"],
            bar0_size       = core_config["phy_bar0_size"])

        # PCIe Endpoint ----------------------------------------------------------------------------
        self.submodules.pcie_endpoint = LitePCIeEndpoint(self.pcie_phy,
            endianness           = core_config["endianness"],
            max_pending_requests = core_config.get("ep_max_pending_requests", 4))

        # PCIe Wishbone Master ---------------------------------------------------------------------
        pcie_wishbone_master = LitePCIeWishboneMaster(self.pcie_endpoint, qword_aligned=core_config["qword_aligned"])
        self.submodules += pcie_wishbone_master
        self.add_wb_master(pcie_wishbone_master.wishbone)

        # PCIe MMAP Master -------------------------------------------------------------------------
        if core_config.get("mmap", False):
            mmap_base        = core_config["mmap_base"]
            mmap_size        = core_config["mmap_size"]
            mmap_translation = core_config.get("mmap_translation", 0x00000000)
            wb = wishbone.Interface(data_width=32)
            self.mem_map["mmap"] = mmap_base
            self.add_wb_slave(mmap_base, wb, mmap_size)
            self.add_memory_region("mmap", mmap_base, mmap_size, type="io")
            axi = AXILiteInterface(data_width=32, address_width=32)
            wb2axi = Wishbone2AXILite(wb, axi, base_address=-mmap_translation)
            self.submodules += wb2axi
            platform.add_extension(axi.get_ios("mmap_axi_lite"))
            axi_pads = platform.request("mmap_axi_lite")
            self.comb += axi.connect_to_pads(axi_pads, mode="master")

        # PCIe MMAP Slave --------------------------------------------------------------------------
        if core_config.get("mmap_slave", False):
            # AXI-Full
            if core_config.get("mmap_slave_axi_full", False):
                pcie_axi_slave = LitePCIeAXISlave(self.pcie_endpoint, data_width=128)
                self.submodules += pcie_axi_slave
                platform.add_extension(pcie_axi_slave.axi.get_ios("mmap_slave_axi"))
                axi_pads = platform.request("mmap_slave_axi")
                self.comb += pcie_axi_slave.axi.connect_to_pads(axi_pads, mode="slave")
            # AXI-Lite
            else:
                platform.add_extension(axi.get_ios("mmap_slave_axi_lite"))
                axi_pads = platform.request("mmap_slave_axi_lite")
                wb = wishbone.Interface(data_width=32)
                axi = AXILiteInterface(data_width=32, address_width=32)
                self.comb += axi.connect_to_pads(axi_pads, mode="slave")
                axi2wb = AXILite2Wishbone(axi, wb)
                self.submodules += axi2wb
                pcie_wishbone_slave = LitePCIeWishboneSlave(self.pcie_endpoint, qword_aligned=core_config["qword_aligned"])
                self.submodules += pcie_wishbone_slave
                self.comb += wb.connect(pcie_wishbone_slave.wishbone)

        # PCIe DMA ---------------------------------------------------------------------------------
        pcie_dmas = []
        self.add_constant("DMA_CHANNELS", core_config["dma_channels"])
        for i in range(core_config["dma_channels"]):
            pcie_dma = LitePCIeDMA(self.pcie_phy, self.pcie_endpoint,
                with_buffering    = core_config["dma_buffering"] != 0,
                buffering_depth   = core_config["dma_buffering"],
                with_loopback     = core_config["dma_loopback"],
                with_synchronizer = core_config["dma_synchronizer"],
                with_monitor      = core_config["dma_monitor"])
            pcie_dma = stream.BufferizeEndpoints({"sink"   : stream.DIR_SINK})(pcie_dma)
            pcie_dma = stream.BufferizeEndpoints({"source" : stream.DIR_SOURCE})(pcie_dma)
            setattr(self.submodules, "pcie_dma" + str(i), pcie_dma)
            self.add_csr("pcie_dma{}".format(i))
            dma_writer_ios = platform.request("dma{}_writer_axi".format(i))
            dma_reader_ios = platform.request("dma{}_reader_axi".format(i))
            self.comb += [
                # Writer IOs
                pcie_dma.sink.valid.eq(dma_writer_ios.tvalid),
                dma_writer_ios.tready.eq(pcie_dma.sink.ready),
                pcie_dma.sink.last.eq(dma_writer_ios.tlast),
                pcie_dma.sink.data.eq(dma_writer_ios.tdata),
                pcie_dma.sink.first.eq(dma_writer_ios.tuser),

                # Reader IOs
                dma_reader_ios.tvalid.eq(pcie_dma.source.valid),
                pcie_dma.source.ready.eq(dma_reader_ios.tready),
                dma_reader_ios.tlast.eq(pcie_dma.source.last),
                dma_reader_ios.tdata.eq(pcie_dma.source.data),
                dma_reader_ios.tuser.eq(pcie_dma.source.first),
            ]

        # PCIe MSI ---------------------------------------------------------------------------------
        if core_config.get("msi_x", False):
            assert core_config["msi_irqs"] <= 32
            self.submodules.pcie_msi = LitePCIeMSIX(self.pcie_endpoint, width=64)
            self.comb += self.pcie_msi.irqs[32:32+core_config["msi_irqs"]].eq(platform.request("msi_irqs"))
        else:
            assert core_config["msi_irqs"] <= 16
            if core_config.get("msi_multivector", False):
                self.submodules.pcie_msi = LitePCIeMSIMultiVector(width=32)
            else:
                self.submodules.pcie_msi = LitePCIeMSI(width=32)
            self.comb += self.pcie_msi.source.connect(self.pcie_phy.msi)
            self.comb += self.pcie_msi.irqs[16:16+core_config["msi_irqs"]].eq(platform.request("msi_irqs"))
        self.interrupts = {}
        for i in range(core_config["dma_channels"]):
            self.interrupts["pcie_dma" + str(i) + "_writer"] = getattr(self, "pcie_dma" + str(i)).writer.irq
            self.interrupts["pcie_dma" + str(i) + "_reader"] = getattr(self, "pcie_dma" + str(i)).reader.irq
        for i, (k, v) in enumerate(sorted(self.interrupts.items())):
            self.comb += self.pcie_msi.irqs[i].eq(v)
            self.add_constant(k.upper() + "_INTERRUPT", i)
        assert len(self.interrupts.keys()) <= 16

    def generate_documentation(self, build_name, **kwargs):
        from litex.soc.doc import generate_docs
        generate_docs(self, "documentation".format(build_name),
            project_name = "LitePCIe standalone core",
            author       = "Enjoy-Digital")
        os.system("sphinx-build -M html documentation/ documentation/_build".format(build_name, build_name))

# Build --------------------------------------------------------------------------------------------

def main():
    parser = argparse.ArgumentParser(description="LitePCIe standalone core generator")
    parser.add_argument("config", help="YAML config file")
    parser.add_argument("--doc",  action="store_true", help="Build documentation")
    args = parser.parse_args()
    core_config = yaml.load(open(args.config).read(), Loader=yaml.Loader)

    # Convert YAML elements to Python/LiteX --------------------------------------------------------
    for k, v in core_config.items():
        replaces = {"False": False, "True": True, "None": None}
        for r in replaces.keys():
            if v == r:
                core_config[k] = replaces[r]

    # Generate core --------------------------------------------------------------------------------
    if core_config["phy"]  == "C5PCIEPHY":
        from litex.build.altera import AlteraPlatform
        from litepcie.phy.c5pciephy import C5PCIEPHY
        platform = AlteraPlatform("", io=[])
        core_config["phy"]           = C5PCIEPHY
        core_config["qword_aligned"] = True
        core_config["endianness"]    = "little"
    elif core_config["phy"] == "S7PCIEPHY":
        from litex.build.xilinx import XilinxPlatform
        from litepcie.phy.s7pciephy import S7PCIEPHY
        platform = XilinxPlatform(core_config["phy_device"], io=[], toolchain="vivado")
        core_config["phy"]           = S7PCIEPHY
        core_config["qword_aligned"] = False
        core_config["endianness"]    = "big"
    elif core_config["phy"] == "USPCIEPHY":
        from litex.build.xilinx import XilinxPlatform
        from litepcie.phy.uspciephy import USPCIEPHY
        platform = XilinxPlatform(core_config["phy_device"], io=[], toolchain="vivado")
        core_config["phy"]           = USPCIEPHY
        core_config["qword_aligned"] = False
        core_config["endianness"]    = "little"
    elif core_config["phy"] == "USPPCIEPHY":
        from litex.build.xilinx import XilinxPlatform
        from litepcie.phy.usppciephy import USPPCIEPHY
        platform = XilinxPlatform(core_config["phy_device"], io=[], toolchain="vivado")
        core_config["phy"]           = USPPCIEPHY
        core_config["qword_aligned"] = False
        core_config["endianness"]    = "little"
    else:
        raise ValueError("Unsupported PCIe PHY: {}".format(core_config["phy"]))
    soc      = LitePCIeCore(platform, core_config)
    builder  = Builder(soc, output_dir="build", compile_gateware=False)
    builder.build(build_name="litepcie_core", regular_comb=True)
    generate_litepcie_software_headers(soc, "./")

    if args.doc:
        soc.generate_documentation("litepcie_core")

if __name__ == "__main__":
    main()



#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2018 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.gen import *

from litex.soc.interconnect import stream
from litex.soc.interconnect.stream import *
from litex.soc.interconnect.packet import *

# Constants/Helpers --------------------------------------------------------------------------------

KB = 1024
MB = 1024*KB
GB = 1024*MB

def get_bar_mask(size):
    mask = 0
    found = 0
    for i in range(32):
        if size%2:
            found = 1
        if found:
            mask |= (1 << i)
        size = size >> 1
    return mask

# Layouts ------------------------------------------------------------------------------------------

def phy_layout(data_width):
    layout = [
        ("dat", data_width),
        ("be",  data_width//8)
    ]
    return EndpointDescription(layout)

def request_layout(data_width):
    layout = [
        ("we",               1),
        ("adr",             32),
        ("len",             10),
        ("req_id",          16),
        ("tag",              8),
        ("dat",     data_width),
        ("channel",          8), # For routing.
        ("user_id",          8)  # For packet identification.
    ]
    return EndpointDescription(layout)

def completion_layout(data_width):
    layout = [
        ("adr",             32),
        ("len",             10),
        ("end",              1),
        ("req_id",          16),
        ("cmp_id",          16),
        ("err",              1),
        ("tag",              8),
        ("dat",     data_width),
        ("channel",          8), # For routing.
        ("user_id",          8)  # For packet identification.
    ]
    return EndpointDescription(layout)

def msi_layout():
    return [("dat", 8)]

def dma_layout(data_width):
    layout = [("data", data_width)]
    return EndpointDescription(layout)
